#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>

#include <device_io.h>
#include <device.h>

#include <bm.h>

extern block_t* ntfsBlock[3];

typedef struct
{
	block_t *blk;
	off_t offset;
} ntfs_privat_t;

int ntfs_fsync(int fildes)
{
	return 0;
}

/**
 * ntfs_device_unix_io_open - Open a device and lock it exclusively
 * @dev:
 * @flags:
 */
int ntfs_device_unix_io_open(struct ntfs_device *dev, int flags)
{
	if (NDevOpen(dev))
		return -1;

	int fsn = dev->d_name[4] - '0';
	if (fsn < 0 || fsn >= 3)
		return -1;

	NDevSetBlock(dev);
	dev->d_private = (ntfs_privat_t*)malloc(sizeof(ntfs_privat_t));
	if (!dev->d_private)
		return -1;

	((ntfs_privat_t*)dev->d_private)->blk = ntfsBlock[fsn];
	((ntfs_privat_t*)dev->d_private)->offset = 0;

	if ((flags&O_RDWR) != O_RDWR)
		NDevSetReadOnly(dev);

	NDevSetOpen(dev);
	return 0;
}

/**
 * ntfs_device_unix_io_close - Close the device, releasing the lock
 * @dev:
 */
int ntfs_device_unix_io_close(struct ntfs_device *dev)
{
	if (!NDevOpen(dev))
		return -1;

	NDevClearOpen(dev);
	if (dev->d_private)
		free(dev->d_private);
	dev->d_private = NULL;
	return 0;
}

/**
 * ntfs_device_unix_io_seek - Seek to a place on the device
 * @dev:
 * @offset:
 * @whence:
 */
s64 ntfs_device_unix_io_seek(struct ntfs_device *dev, s64 offset, int whence)
{
	ntfs_privat_t *priv = (ntfs_privat_t*)dev->d_private;
	
	if (whence == SEEK_CUR)
		priv->offset += offset;
	else if (whence == SEEK_SET)
		priv->offset = offset;
	else if (whence == SEEK_END)
	{
		uint64_t size;
		if (!priv->blk->ioctl(priv->blk, IOCTL_GET_SIZE, &size))
			priv->offset = size + offset;
	}

	return priv->offset;
}

/**
 * ntfs_device_unix_io_read - Read from the device, from the current location
 * @dev:
 * @buf:
 * @count:
 */
s64 ntfs_device_unix_io_read(struct ntfs_device *dev, void *buf, s64 count)
{
	ntfs_privat_t *priv = (ntfs_privat_t*)dev->d_private;
	ssize_t read = priv->blk->pread(priv->blk, buf, count, priv->offset);
	if (read > 0)
		priv->offset += read;
	return read;
}

/**
 * ntfs_device_unix_io_write - Write to the device, at the current location
 * @dev:
 * @buf:
 * @count:
 */
s64 ntfs_device_unix_io_write(struct ntfs_device *dev, const void *buf, s64 count)
{
	ntfs_privat_t *priv = (ntfs_privat_t*)dev->d_private;
	if (NDevReadOnly(dev)) {
		errno = EROFS;
		return -1;
	}
	ssize_t write = priv->blk->pwrite(priv->blk, buf, count, priv->offset);
	if (write > 0)
		priv->offset += write;
	return write;
}

/**
 * ntfs_device_unix_io_pread - Perform a positioned read from the device
 * @dev:
 * @buf:
 * @count:
 * @offset:
 */
s64 ntfs_device_unix_io_pread(struct ntfs_device *dev, void *buf, s64 count, s64 offset)
{
	ntfs_privat_t *priv = (ntfs_privat_t*)dev->d_private;
	return priv->blk->pread(priv->blk, buf, count, offset);
}

/**
 * ntfs_device_unix_io_pwrite - Perform a positioned write to the device
 * @dev:
 * @buf:
 * @count:
 * @offset:
 */
s64 ntfs_device_unix_io_pwrite(struct ntfs_device *dev, const void *buf,
		s64 count, s64 offset)
{
	ntfs_privat_t *priv = (ntfs_privat_t*)dev->d_private;
	if (NDevReadOnly(dev)) {
		errno = EROFS;
		return -1;
	}
	return priv->blk->pwrite(priv->blk, buf, count, offset);
}

/**
 * ntfs_device_unix_io_sync - Flush any buffered changes to the device
 * @dev:
 */
int ntfs_device_unix_io_sync(struct ntfs_device *dev)
{
	return 0;
}

/**
 * ntfs_device_unix_io_stat - Get information about the device
 * @dev:
 * @buf:
 */
int ntfs_device_unix_io_stat(struct ntfs_device *dev, struct stat *buf)
{
	return -1;
}

/**
 * ntfs_device_unix_io_ioctl - Perform an ioctl on the device
 * @dev:
 * @request:
 * @argp:
 */
int ntfs_device_unix_io_ioctl(struct ntfs_device *dev, int request,
		void *argp)
{
	ntfs_privat_t *priv = (ntfs_privat_t*)dev->d_private;
	
    // Figure out which i/o control was requested
    switch (request) {

        // Get block device size (sectors)
        #if defined(BLKGETSIZE)
        case BLKGETSIZE: {
            return priv->blk->ioctl(priv->blk, IOCTL_GET_SECTOR_COUNT, argp);
        }
        #endif

        // Get block device size (bytes)
        #if defined(BLKGETSIZE64)
        case BLKGETSIZE64: {
			return priv->blk->ioctl(priv->blk, IOCTL_GET_SIZE, argp);
        }
        #endif
/*
        // Get hard drive geometry
        #if defined(HDIO_GETGEO)
        case HDIO_GETGEO: {
            struct hd_geometry *geo = (struct hd_geometry*)argp;
            geo->sectors = 0;
            geo->heads = 0;
            geo->cylinders = 0;
            geo->start = 0;
            return -1;
        }
        #endif
*/
        // Get block device sector size (bytes)
        #if defined(BLKSSZGET)
        case BLKSSZGET: {
			uint32_t sz = 512;
			int ret = priv->blk->ioctl(priv->blk, IOCTL_GET_SECTOR_SIZE, &sz);
            *(int*)argp = sz;
            return ret;
        }
        #endif

        // Unimplemented ioctrl
        default: {
            ntfs_log_perror("Unimplemented ioctrl %i\n", request);
            errno = EOPNOTSUPP;
            return -1;
        }

    }

    return 0;
}

/**
 * Device operations for working with unix style devices and files.
 */
struct ntfs_device_operations ntfs_device_unix_io_ops = {
	.open		= ntfs_device_unix_io_open,
	.close		= ntfs_device_unix_io_close,
	.seek		= ntfs_device_unix_io_seek,
	.read		= ntfs_device_unix_io_read,
	.write		= ntfs_device_unix_io_write,
	.pread		= ntfs_device_unix_io_pread,
	.pwrite		= ntfs_device_unix_io_pwrite,
	.sync		= ntfs_device_unix_io_sync,
	.stat		= ntfs_device_unix_io_stat,
	.ioctl		= ntfs_device_unix_io_ioctl,
};
