/*
 * NTFS FS support
 * 
 * Copyright (c) 2021 Ant-ON
 *
 */

#include <sys/param.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include <fs.h>
#include <bm.h>

#include <ntfs.h>
#include <ntfsfile.h>
#include <ntfsdir.h>
#include <ntfstime.h>

block_t* ntfsBlock[3] = {0};

static ssize_t ntfs_f_read(void *vfd, void *buf, size_t len)
{
	fd_t *fd = (fd_t *)vfd;
	ntfs_file_state *file = (ntfs_file_state*)fd->data;
    return ntfs_read_r(file, buf, len);
}

static ssize_t ntfs_f_write(void *vfd, const void *buf, size_t len)
{
	fd_t *fd = (fd_t *)vfd;
	ntfs_file_state *file = (ntfs_file_state*)fd->data;
    return ntfs_write_r(file, buf, len);
}

static off_t ntfs_f_seek(void *vfd, off_t offset, int dir)
{
	fd_t *fd = (fd_t *)vfd;
	ntfs_file_state *file = (ntfs_file_state*)fd->data;
    return ntfs_seek_r(file, offset, dir);
}

static void ntfs_f_close(void *vfd)
{
	fd_t *fd = (fd_t *)vfd;
	if (!fd->data)
		return;

	ntfs_file_state *file = (ntfs_file_state*)fd->data;
	fd->data = NULL;

	ntfs_close_r(file);
	free(file);
	//free(fd); - free in FS.c
}

static void* ntfs_fs_fopen(void *vfs, const char *path, int flags)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;

	ntfs_file_state *file = (ntfs_file_state*)malloc(sizeof(ntfs_file_state));
	if (ntfs_open_r(vol, file, path, flags, 0777))
	{
		ntfs_log_error("open file failed %d (%s)\n", errno, strerror(errno));
		free(file);
		return NULL;
	}

	fd_t *fd = malloc(sizeof(fd_t));
	fd->read = ntfs_f_read;
	fd->write = ntfs_f_write;
	fd->seek = ntfs_f_seek;
	fd->close = ntfs_f_close;
	fd->data = file;
	return fd;
}

static int ntfs_fs_mkdir(void *vfs, const char *path)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;
	return ntfs_mkdir_r(vol, path, 0777);
}

static int ntfs_fs_rename(void *vfs, const char *path, const char *pathNew)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;
	return ntfs_rename_r(vol, path, pathNew);
}

static int ntfs_fs_remove(void *vfs, const char *path)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;
	return ntfs_unlink_r(vol, path);
}

static int ntfs_fs_list(void *vfs, void *ctx, const char *path, _filter filter)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;

	ntfs_dir_state dir;
    if (ntfs_diropen_r(vol, &dir, path))
    {
		char name[255];
		struct stat st;
		while (!ntfs_dirnext_r(&dir, name, &st))
		{
			int flags = (st.st_mode&S_IFDIR)?F_FLAG_DIR:0;
			filter(ctx, name, st.st_size, ntfs2timespec(st.st_mtime).tv_sec, flags);
		}

		ntfs_dirclose_r(&dir);
		return 0;
	}

	return -1;
}

static int ntfs_fs_stat(void *vfs, const char *path, stat_t *stat)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;

	struct stat st;
    int r = ntfs_stat_r(vol, path, &st);
	if (!r)
	{
		stat->time = st.st_atime;
		stat->size = st.st_size;
		stat->flags = (st.st_mode&S_IFDIR)?F_FLAG_DIR:0;
	}

	return r;
}

static int ntfs_fs_statfs(void *vfs, stat_fs_t *stat_fs)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;
	
	struct statvfs st;
	int r = ntfs_statvfs_r(vol, &st);
	if (!r)
	{
		stat_fs->name = strdup(vol->vol_name);
		stat_fs->size = (int64_t)st.f_blocks * st.f_bsize;
		stat_fs->free = (int64_t)st.f_bfree * st.f_bsize;
		stat_fs->flags = 0;
	}

	return r;
}

static void ntfs_fs_fsfree(void *vfs)
{
	fs_t *fs = (fs_t *)vfs;
	ntfs_volume *vol = (ntfs_volume*)fs->data;

	if (vol)
		ntfs_umount(vol, 1);
	free(fs);
}

#ifdef MODULE_FS
void* __UI fs_open(void *block, int flags)
#else
void* __UI ntfs_fs_open(void *block, int flags)
#endif
{
	ntfs_volume *vol = NULL;

	int fsn;
	for (fsn=0; fsn<3; fsn++)
		if (!ntfsBlock[fsn])
		{
			ntfsBlock[fsn] = (block_t*)block;
			break;
		}
	if (fsn >= 3)
		return NULL;
	
	char path[] = "/dev_";
	path[4] = '0' + fsn;

	vol = ntfs_mount(path, (false&FLAG_FS_OPEN_RO)?NTFS_MNT_RDONLY:0);
    if (!vol)
		goto error;

	fs_t *fs = malloc(sizeof(fs_t));
	fs->open = ntfs_fs_fopen;
	fs->mkdir = ntfs_fs_mkdir;
	fs->rename = ntfs_fs_rename;
	fs->remove = ntfs_fs_remove;
	fs->list = ntfs_fs_list;
	fs->stat = ntfs_fs_stat;
	fs->statfs = ntfs_fs_statfs;
	fs->fsfree = ntfs_fs_fsfree;
	fs->data = vol;

	ntfsBlock[fsn] = NULL;
	return fs;

error:
	if (vol)
		ntfs_umount(vol, 1);

	ntfsBlock[fsn] = NULL;
	return NULL;
}
