/**
 * ntfsinternal.h - Internal support routines for NTFS-based devices.
 *
 * Copyright (c) 2010 Dimok
 * Copyright (c) 2009 Rhys "Shareese" Koedijk
 * Copyright (c) 2006 Michael "Chishm" Chisholm
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "ntfs.h"
#include "ntfsdir.h"
#include "ntfsfile.h"

ntfs_inode *ntfsOpenEntry (ntfs_volume *vol, const char *path)
{
    return ntfsParseEntry(vol, path, 1);
}

ntfs_inode *ntfsParseEntry (ntfs_volume *vol, const char *path, int reparseLevel)
{
    ntfs_inode *ni = NULL;
    char *target = NULL;
    int attr_size;

    // Sanity check
    if (!vol) {
        errno = ENODEV;
        return NULL;
    }

    // Get the actual path of the entry
    if (path[0] != '/')
        return NULL;

    // Find the entry, taking into account our current directory (if any)
    ni = ntfs_pathname_to_inode(vol, NULL, path);

    // If the entry was found and it has reparse data then parse its true path;
    // this resolves the true location of symbolic links and directory junctions
    if (ni && (ni->flags & FILE_ATTR_REPARSE_POINT)) {
        if (ntfs_possible_symlink(ni)) {

            // Sanity check, give up if we are parsing to deep
            if (reparseLevel > NTFS_MAX_SYMLINK_DEPTH) {
                ntfsCloseEntry(ni);
                errno = ELOOP;
                return NULL;
            }

            // Get the target path of this entry
            target = ntfs_make_symlink(ni, path, &attr_size);
            if (!target) {
                ntfsCloseEntry(ni);
                return NULL;
            }

            // Close the entry (we are no longer interested in it)
            ntfsCloseEntry(ni);

            // Parse the entries target
            ni = ntfsParseEntry(vol, target, reparseLevel++);

            // Clean up
            // use free because the value was not allocated with malloc
            free(target);

        }
    }

    return ni;
}

void ntfsCloseEntry (ntfs_inode *ni)
{
    // Sync the entry (if it is dirty)
    if (NInoDirty(ni))
        ntfsSync(ni);

    // Close the entry
    ntfs_inode_close(ni);
}


ntfs_inode *ntfsCreate(ntfs_volume *vol, const char *path, mode_t type, const char *target)
{
    ntfs_inode *dir_ni = NULL, *ni = NULL;
    char *dir = NULL;
    char *name = NULL;
    ntfschar *uname = NULL, *utarget = NULL;
    int uname_len, utarget_len;

    // Sanity check
    if (!vol) {
        errno = ENODEV;
        return NULL;
    }
    if (!path) {
        errno = EINVAL;
        return NULL;
    }

    // Get the unicode name for the entry and find its parent directory
    // TODO: This looks horrible, clean it up
    dir = strdup(path);
    if (!dir) {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if (name)
        name++;
    else
        name = dir;
    uname_len = ntfsLocalToUnicode(name, &uname);
    if (uname_len < 0) {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if(name)
    {
        name++;
        name[0] = 0;
    }

    // Open the entries parent directory
    dir_ni = ntfsOpenEntry(vol, dir);
    if (!dir_ni) {
        goto cleanup;
    }

    // Create the entry
    switch (type) {

        // Symbolic link
        case S_IFLNK:
            if (!target) {
                errno = EINVAL;
                goto cleanup;
            }
            utarget_len = ntfsLocalToUnicode(target, &utarget);
            if (utarget_len < 0) {
                errno = EINVAL;
                goto cleanup;
            }
            ni = ntfs_create_symlink(dir_ni, 0, uname, uname_len,  utarget, utarget_len);
            break;

        // Directory or file
        case S_IFDIR:
        case S_IFREG:
            ni = ntfs_create(dir_ni, 0, uname, uname_len, type);
            break;

    }

    // If the entry was created
    if (ni) {

        // Mark the entry for archiving
        ni->flags |= FILE_ATTR_ARCHIVE;

        // Mark the entry as dirty
        NInoSetDirty(ni);

        // Sync the entry to disc
        ntfsSync(ni);

        // Update parent directories times
        ntfsUpdateTimes(dir_ni, NTFS_UPDATE_MCTIME);

    }

cleanup:

    if(dir_ni)
        ntfsCloseEntry(dir_ni);
    if(utarget)
        free(utarget);
    if(uname)
        free(uname);
    if(dir)
        free(dir);
    return ni;
}

int ntfsLink (ntfs_volume *vol, const char *old_path, const char *new_path)
{
    ntfs_inode *dir_ni = NULL, *ni = NULL;
    char *dir = NULL;
    char *name = NULL;
    ntfschar *uname = NULL;
    int uname_len;
    int res = 0;

    // Sanity check
    if (!vol) {
        errno = ENODEV;
        return -1;
    }
    if (!old_path || !new_path) {
        errno = EINVAL;
        return -1;
    }

    // Get the unicode name for the entry and find its parent directory
    // TODO: This looks horrible, clean it up
    dir = strdup(new_path);
    if (!dir) {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if (name)
        name++;
    else
        name = dir;
    uname_len = ntfsLocalToUnicode(name, &uname);
    if (uname_len < 0) {
        errno = EINVAL;
        goto cleanup;
    }
    *name = 0;

    // Find the entry
    ni = ntfsOpenEntry(vol, old_path);
    if (!ni) {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Open the entries new parent directory
    dir_ni = ntfsOpenEntry(vol, dir);
    if (!dir_ni) {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Link the entry to its new parent
    if (ntfs_link(ni, dir_ni, uname, uname_len)) {
        res = -1;
        goto cleanup;
    }

    // Update entry times
    ntfsUpdateTimes(dir_ni, NTFS_UPDATE_MCTIME);

    // Sync the entry to disc
    ntfsSync(ni);

cleanup:

    if(dir_ni)
        ntfsCloseEntry(dir_ni);
    if(ni)
        ntfsCloseEntry(ni);
    if(uname)
        free(uname);
    if(dir)
        free(dir);
    return res;
}

int ntfsUnlink (ntfs_volume *vol, const char *path)
{
    ntfs_inode *dir_ni = NULL, *ni = NULL;
    char *dir = NULL;
    char *name = NULL;
    ntfschar *uname = NULL;
    int uname_len;
    int res = 0;

    // Sanity check
    if (!vol) {
        errno = ENODEV;
        return -1;
    }
    if (!path) {
        errno = EINVAL;
        return -1;
    }

    // Get the unicode name for the entry and find its parent directory
    // TODO: This looks horrible
    dir = strdup(path);
    if (!dir) {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if (name)
        name++;
    else
        name = dir;
    uname_len = ntfsLocalToUnicode(name, &uname);
    if (uname_len < 0) {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if(name)
    {
        name++;
        name[0] = 0;
    }

    // Find the entry
    ni = ntfsOpenEntry(vol, path);
    if (!ni) {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Open the entries parent directory
    dir_ni = ntfsOpenEntry(vol, dir);
    if (!dir_ni) {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Unlink the entry from its parent
    if (ntfs_delete(vol, path, ni, dir_ni, uname, uname_len)) {
        res = -1;
    }

    // ntfs_delete() ALWAYS closes ni and dir_ni; so no need for us to anymore
    dir_ni = ni = NULL;

cleanup:

    if(dir_ni)
        ntfsCloseEntry(dir_ni);
    if(ni)
        ntfsCloseEntry(ni);
    if(uname)
        free(uname);
    if(dir)
        free(dir);
    return res;
}

int ntfsSync(ntfs_inode *ni)
{
    // Sanity check
    if (!ni) {
        errno = ENOENT;
        return -1;
    }

    // Sync the entry
    return ntfs_inode_sync(ni);

}

int ntfsStat (ntfs_inode *ni, struct stat *st)
{
    ntfs_attr *na = NULL;
    int res = 0;

    // Sanity check
    if (!ni) {
        errno = ENOENT;
        return -1;
    }

    // Short circuit cases were we don't actually have to do anything
    if (!st)
        return 0;

    // Zero out the stat buffer
    memset(st, 0, sizeof(struct stat));

    // Is this entry a directory
    if (ni->mrec->flags & MFT_RECORD_IS_DIRECTORY) {
        st->st_mode = S_IFDIR | 0777;
        st->st_nlink = 1;

        // Open the directories index allocation table attribute
        na = ntfs_attr_open(ni, AT_INDEX_ALLOCATION, NTFS_INDEX_I30, 4);
        if (na) {
            st->st_size = na->data_size;
            st->st_blocks = na->allocated_size >> 9;
            ntfs_attr_close(na);
        }

    // Else it must be a file
    } else {
        st->st_mode = S_IFREG | 0777;
        st->st_size = ni->data_size;
        st->st_blocks = (ni->allocated_size + 511) >> 9;
        st->st_nlink = le16_to_cpu(ni->mrec->link_count);
    }

    // Fill in the generic entry stats
    st->st_ino = ni->mft_no;
    st->st_atime = ni->last_access_time;
    st->st_ctime = ni->last_mft_change_time;
    st->st_mtime = ni->last_data_change_time;

    // Update entry times
    ntfsUpdateTimes(ni, NTFS_UPDATE_ATIME);

    return res;
}

void ntfsUpdateTimes (ntfs_inode *ni, ntfs_time_update_flags mask)
{
#ifdef ATIME_DISABLED	
    // Run the access time update strategy against the device driver settings first
    mask &= ~NTFS_UPDATE_ATIME;
#endif // ATIME_DISABLED

    // Update entry times
    if (ni && mask)
        ntfs_inode_update_times(ni, mask);

    return;
}

int ntfsUnicodeToLocal (const ntfschar *ins, const int ins_len, char **outs, int outs_len)
{
    int len = 0;
    int i;

    // Sanity check
    if (!ins || !ins_len || !outs)
        return 0;

    char * ucstombs_out = NULL;

    // Convert the unicode string to our current local
    len = ntfs_ucstombs(ins, ins_len, &ucstombs_out, outs_len);

    if(ucstombs_out)
    {
        //use proper allocation
        *outs = (char *) malloc(strlen(ucstombs_out) + 1);
        if(!*outs)
        {
            errno = ENOMEM;
            return -1;
        }
        strcpy(*outs, ucstombs_out);
        free(ucstombs_out);
        ucstombs_out = NULL;
    }

    if (len == -1 && errno == EILSEQ)
    {
        // The string could not be converted to the current local,
        // do it manually by replacing non-ASCII characters with underscores
        if (!*outs || outs_len >= ins_len)
        {
            if (!*outs)
            {
                *outs = (char *) malloc(ins_len + 1);
                if (!*outs) {
                    errno = ENOMEM;
                    return -1;
                }
            }
            for (i = 0; i < ins_len; i++) {
                ntfschar uc = le16_to_cpu(ins[i]);
                if (uc > 0xff)
                    uc = (ntfschar)'_';
                *outs[i] = (char)uc;
            }
            *outs[ins_len] = (ntfschar)'\0';
            len = ins_len;
        }
    }

    return len;
}

int ntfsLocalToUnicode (const char *ins, ntfschar **outs)
{
    // Sanity check
    if (!ins || !outs)
        return 0;

    // Convert the local string to unicode
    return ntfs_mbstoucs(ins, outs);
}
