/**
 * ntfsinternal.h - Internal support routines for NTFS-based devices.
 *
 * Copyright (c) 2009 Rhys "Shareese" Koedijk
 * Copyright (c) 2006 Michael "Chishm" Chisholm
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _NTFSINTERNAL_H
#define _NTFSINTERNAL_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdbool.h> 

#include "types.h"
#include "compat.h"
#include "logging.h"
#include "layout.h"
#include "device.h"
#include "volume.h"
#include "dir.h"
#include "inode.h"
#include "attrib.h"
#include "reparse.h"
#include "security.h"
#include "efs.h"
#include "unistr.h"

#define NTFS_MOUNT_PREFIX                   "ntfs" /* Device name prefix to use when auto-mounting */
#define NTFS_MAX_PARTITIONS                 32 /* Maximum number of partitions that can be found */
#define NTFS_MAX_MOUNTS                     10 /* Maximum number of mounts available at one time */
#define NTFS_MAX_SYMLINK_DEPTH              10 /* Maximum search depth when resolving symbolic links */

#define NTFS_OEM_ID                         cpu_to_le64(0x202020205346544eULL) /* "NTFS    " */

#define MBR_SIGNATURE                       cpu_to_le16(0xAA55)
#define EBR_SIGNATURE                       cpu_to_le16(0xAA55)

#define PARTITION_STATUS_NONBOOTABLE        0x00 /* Non-bootable */
#define PARTITION_STATUS_BOOTABLE           0x80 /* Bootable (active) */

#define PARTITION_TYPE_EMPTY                0x00 /* Empty */
#define PARTITION_TYPE_DOS33_EXTENDED       0x05 /* DOS 3.3+ extended partition */
#define PARTITION_TYPE_NTFS                 0x07 /* Windows NT NTFS */
#define PARTITION_TYPE_WIN95_EXTENDED       0x0F /* Windows 95 extended partition */

/* Forward declarations */
struct _ntfs_file_state;
struct _ntfs_dir_state;

/* Miscellaneous helper/support routines */
ntfs_inode *ntfsOpenEntry (ntfs_volume *vol, const char *path);
ntfs_inode *ntfsParseEntry (ntfs_volume *vol, const char *path, int reparseLevel);
ntfs_inode *ntfsCreate (ntfs_volume *vol, const char *path, mode_t type, const char *target);

int ntfsLink (ntfs_volume *vol, const char *old_path, const char *new_path);
int ntfsUnlink (ntfs_volume *vol, const char *path);

void ntfsCloseEntry (ntfs_inode *ni);
int ntfsSync (ntfs_inode *ni);
int ntfsStat (ntfs_inode *ni, struct stat *st);
void ntfsUpdateTimes(ntfs_inode *ni, ntfs_time_update_flags mask);

int ntfsUnicodeToLocal (const ntfschar *ins, const int ins_len, char **outs, int outs_len);
int ntfsLocalToUnicode (const char *ins, ntfschar **outs);

#endif /* _NTFSINTERNAL_H */
